# -*- coding: utf-8 -*-
ITERABLE_TYPE_STREAM = 10
ITERABLE_TYPE_FILE = 20
ITERABLE_TYPE_CODEC = 30
DEFAULT_BULK_NUMBER = 100

import io
import typing


class BaseCodec:
    """Basic codec class"""
    def __init__(self, filename: str = None, fileobj: typing.IO = None, mode: str = 'r', open_it: bool = False):
        self._fileobj = fileobj
        self.filename = filename
        self.mode = mode
        if open_it:
            self.open()
        pass

    @staticmethod
    def fileexts():
        """Return file extensions"""
        raise NotImplementedError
    

    def reset(self):
        """Reset file"""
        self._fileobj.seek(0)

    def open(self):
        raise NotImplementedError
 
    def fileobj(self):
        """Return file object"""
        return self._fileobj
    
    def close(self):
        """Close codec. Not implemented by default"""
        raise NotImplementedError

    def textIO(self, encoding:str = 'utf8'):
        """Return text wrapper over binary stream"""
        return io.TextIOWrapper(self.fileobj(), encoding=encoding, write_through=True)


class BaseIterable:
    """Base iterable data class"""
    def __init__(self):
        pass

    def reset(self):
        """Reset iterator"""
        raise NotImplementedError

    @staticmethod
    def id():
        """Identifier of selected destination"""
        raise NotImplementedError

    def read(self, skip_empty:bool = True):
        """Read single record"""
        raise NotImplementedError

    def read_bulk(self, num:int = DEFAULT_BULK_NUMBER):
        """Read multiple records"""
        raise NotImplementedError

    @staticmethod
    def is_flatonly():
        """Is source flat by only. Default: False"""
        return False

    def is_flat(self):
        """Is source flat. Default: """ 
        if self.__class__().is_flatonly():
            return True
        raise NotImplementedError

    def is_streaming(self):
        """Is source streaming. Default: False"""
        return False

    def __next__(self):
        return self.read()

    def __iter__(self):
        self.reset()
        return self

    def write(self, rec):
        """Write single record"""
        raise NotImplementedError

    def write_bulk(self, records):
        """Write multiple records"""
        raise NotImplementedError


class BaseFileIterable(BaseIterable):
    """Basic file iterable"""
    datamode = 'text'

    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, binary:bool = False, encoding:str = 'utf8', noopen:bool = False, mode:str = 'r'):
        """Init basic file iterable"""
        self.filename = filename
        self.noopen = noopen
        self.encoding = encoding
        self.binary = binary
        self.mode = mode
        self.codec = codec
        if stream is not None:
            self.stype = ITERABLE_TYPE_STREAM
        elif filename is not None:
            self.stype = ITERABLE_TYPE_FILE
        elif codec is not None:
            self.stype = ITERABLE_TYPE_CODEC
        self.fobj = None

        if self.stype == ITERABLE_TYPE_FILE:
            if not noopen:                
                self.open()
        elif self.stype == ITERABLE_TYPE_STREAM:
            self.fobj = stream
        elif self.stype == ITERABLE_TYPE_CODEC:
            if not noopen:                
                self.fobj = self.codec.open() 
                if self.datamode == 'text':
                    self.fobj = self.codec.textIO(encoding=self.encoding) 
                  

    def open(self):
        """Open file as file data source"""
        if self.stype ==  ITERABLE_TYPE_FILE:
            self.fobj = open(self.filename, self.mode + 'b') if self.binary else open(self.filename, self.mode, encoding=self.encoding)
            return self.fobj
        else:
            raise NotImplementedError
	

    def reset(self):
        """Reset file using seek(0)"""
        if self.stype == ITERABLE_TYPE_FILE:
            if self.fobj is not None:
                self.fobj.seek(0)
        elif self.stype == ITERABLE_TYPE_CODEC:
            if self.fobj is not None and self.mode not in ['w', 'wb']:
                self.fobj.seek(0)
#            self.codec.reset()

    def close(self):
        """Close file as file data source"""
        if self.stype == ITERABLE_TYPE_FILE:
            if self.fobj is not None:
                self.fobj.close()
        elif self.stype == ITERABLE_TYPE_CODEC:
            if self.codec is not None:
                self.codec.close()



from __future__ import annotations
import typing
from xlrd import open_workbook
import xlrd
from ..base import BaseFileIterable
import datetime

def read_row_keys(rownum, ncols, sheet):
    """Read single row by row num"""
    tmp = list()
    for i in range(0, ncols):
        ct = sheet.cell_type(rownum, i)
        cell_value = sheet.cell_value(rownum, i)
        get_col = str(cell_value)
        tmp.append(get_col)    
    return tmp


def read_single_row(rownum, ncols, datemode, keys, sheet):
    """Read single row by row num"""
    tmp = list()
    for i in range(0, ncols):
        ct = sheet.cell_type(rownum, i)
        cell_value = sheet.cell_value(rownum, i)
        if ct == xlrd.XL_CELL_DATE:
            # Returns a tuple.
            dt_tuple = xlrd.xldate_as_tuple(cell_value, datemode)
            # Create datetime object from this tuple.
            get_col = str(datetime.datetime(             
                dt_tuple[0], dt_tuple[1], dt_tuple[2],
                dt_tuple[3], dt_tuple[4], dt_tuple[5]
            ))
        elif ct == xlrd.XL_CELL_NUMBER:
            get_col = int(cell_value)
        else:
            get_col = str(cell_value)
        tmp.append(get_col)
    row = dict(zip(keys, tmp))
    return row



class XLSIterable(BaseFileIterable):
    datamode = 'binary'
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode='r', keys: list[str] = None, page:int = 0, start_line:int = 0):
        super(XLSIterable, self).__init__(filename, stream, codec=codec, binary=True, mode=mode, noopen=True)
        self.page = page
        self.start_line = start_line
        self.pos = start_line
        self.keys = keys
        self.extracted_keys = keys is None
        self.reset()
        pass

    def reset(self):
        """Reopen file and open sheet"""
        super(XLSIterable, self).reset()
        self.pos = self.start_line
        self.workbook = open_workbook(self.filename)
        self.sheet = self.workbook.sheet_by_index(self.page)
        if self.extracted_keys:
            self.keys = read_row_keys(self.pos, self.sheet.ncols, self.sheet)
            self.pos += 1

    @staticmethod
    def id() -> str:
        """ID of the data source type"""
        return 'xls'

    @staticmethod
    def is_flatonly() -> bool:
        """Flag that data is flat"""
        return True

    def read(self) -> dict:
        """Read single XLS record"""
        if self.pos >= self.sheet.nrows:
            raise StopIteration
        row = read_single_row(self.pos, self.sheet.ncols, self.workbook.datemode, self.keys, self.sheet)
        self.pos += 1
        return row

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk XLS records"""
        chunk = []
        ncols = self.sheet.ncols
        datemode = self.workbook.datemode
        for n in range(0, num):
            if self.pos >= self.sheet.nrows:
                raise StopIteration
            row = read_single_row(self.pos, ncols, datemode, self.keys, self.sheet)
            chunk.append(row)
            self.pos += 1
        return chunk



from __future__ import annotations
import typing
from collections import defaultdict
from collections import OrderedDict
import csv
import chardet
from statistics import mean

from ..base import BaseIterable

DEFAULT_DELIMITERS = [',', ';', '\t', '|']

def detect_encoding(filename:str = None, stream:typing.IO=None, limit:int=1000000) -> str:
    """Detect file or file object encoding reading 1MB data by default and using chardet"""    
    if filename is not None:
        f = open(filename, 'rb')
        chunk = f.read(limit)
        f.close()
        return chardet.detect(chunk)
    else:
        return chardet.detect(stream.read(limit))

def detect_delimiter(filename:str = None, stream:typing.IO=None, encoding:str ='utf8', limit:int=20, threshold=0.6) -> str:
    """Detect CSV file or file object delimiter with known encoding and limit with number of lines"""
    lines = []
    char_map = {}
    if filename:
        f = open(filename, 'r', encoding=encoding)
        for n in range(0, limit):
            line = f.readline().strip()
            if len(line) > 0:
                lines.append(line)
        f.close()
    else:
        for n in range(0, limit):
            lines.append(stream.readline())

    for char in DEFAULT_DELIMITERS:
        char_map[char] = []
    
    for line in lines:
        for char in DEFAULT_DELIMITERS:
            char_map[char].append(line.count(char))
    
    candidates = {}
    for char in char_map:        
        if min(char_map[char]) != 0 and mean(char_map[char]) / max(char_map[char]) > threshold:
            candidates[char] = max(char_map[char])
            
    delimiter = max(candidates, key=candidates.get)
    return delimiter

def get_dict_value(d:dict, keys: list[str]):
    """Return value of selected dict key"""
    out = []
    if d is None:
        return out
#    keys = key.split('.')
    if len(keys) == 1:
        if type(d) == type({}) or isinstance(d, OrderedDict):
            if keys[0] in d.keys():
                out.append(d[keys[0]])
        else:
            for r in d:
                if r and keys[0] in r.keys():
                    out.append(r[keys[0]])
#        return out
    else:
        if type(d) == type({}) or isinstance(d, OrderedDict):
            if keys[0] in d.keys():
                out.extend(get_dict_value(d[keys[0]], keys[1:]))
        else:
            for r in d:
                if keys[0] in r.keys():
                    out.extend(get_dict_value(r[keys[0]], keys[1:]))
    return out


def strip_dict_fields(record, fields, startkey=0):
    """Remove selected dict fields"""
    keys = record.keys()
    localf = []
    for field in fields:
        if len(field) > startkey:
            localf.append(field[startkey])
    for k in list(keys):
        if k not in localf:
            del record[k]

    if len(k) > 0:
        for k in record.keys():
            if type(record[k]) == type({}):
                record[k] = strip_dict_fields(record[k], fields, startkey + 1)
    return record


def dict_generator(indict:dict, pre:list=None):
    """Processes python dictionary and return list of key values
    :param indict
    :param pre
    :return generator"""
    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in list(indict.items()):
            if key == "_id":
                continue
            if isinstance(value, dict):
                #                print 'dgen', value, key, pre
                for d in dict_generator(value, pre + [key]):
                    yield d
            elif isinstance(value, list) or isinstance(value, tuple):
                for v in value:
                    if isinstance(v, dict):
                        #                print 'dgen', value, key, pre
                        for d in dict_generator(v, pre + [key]):
                            yield d
#                    for d in dict_generator(v, [key] + pre):
#                        yield d
            else:
                yield pre + [key, value]
    else:
        yield indict


def guess_int_size(value:int):
    """Guess integer size"""
    if value < 255:
        return 'uint8'
    if value < 65535:
        return 'uint16'
    return 'uint32'

def guess_datatype(s:str, qd:Object) -> dict:
    """Guesses type of data by string provided
    :param s
    :param qd
    :return datatype"""
    attrs = {'base' : 'str'}
#    s = unicode(s)
    if s is None:
       return {'base' : 'empty'}
    if type(s) == type(1):
        return {'base' : 'int'}
    if type(s) == type(1.0):
        return {'base' : 'float'}
    elif type(s) != type(''):
#        print((type(s)))
        return {'base' : 'typed'}
#    s = s.decode('utf8', 'ignore')
    if s.isdigit():
        if s[0] == 0:
            attrs = {'base' : 'numstr'}
        else:
            attrs = {'base' : 'int', 'subtype' : guess_int_size(int(s))}
    else:
        try:
            i = float(s)
            attrs = {'base' : 'float'}
            return attrs
        except ValueError:
            pass
        if qd:
            is_date = False
            res = qd.match(s)
            if res:
                attrs = {'base': 'date', 'pat': res['pattern']}
                is_date = True
            if not is_date:
                if len(s.strip()) == 0:
                    attrs = {'base' : 'empty'}
    return attrs


def count_file_newlines(filename:str = None, stream:typing.IO = None):
    """Counts number of lines in file"""
    def _make_gen(reader):
        while True:
            b = reader(2 ** 16)
            if not b: break
            yield b

    if fname:
        with open(fname, "rb") as f:
            count = sum(buf.count(b"\n") for buf in _make_gen(f.raw.read))
    else:
        sum(stream.count(b"\n") for buf in _make_gen(f.raw.read))
    return count


def get_dict_keys(iterable:list[dict], limit:int=1000) -> list[str]:
    """Returns dictionary keys"""
    n = 0
    keys = []
    for item in iterable:
        if limit and n > limit:
            break
        n += 1
        dk = dict_generator(item)
        for i in dk:
            k = ".".join(i[:-1])
            if k not in keys:
                keys.append(k)
    return keys

def get_iterable_keys(iterable:BaseIterable, limit:int=1000) -> list[str]:
    """Returns BaseIterable object keys"""
    n = 0
    keys = []
    for item in iterable:
        if limit and n > limit:
            break
        n += 1
        dk = dict_generator(item)
        for i in dk:
            k = ".".join(i[:-1])
            if k not in keys:
                keys.append(k)
    return keys

def is_flat_object(item:object) -> bool:
    """Measures if object is flat"""
    for k, v in item.items():
        if isinstance(v, tuple) or isinstance(v, list):
            return False
        elif isinstance(v, dict):
            if not _is_flat(v): return False
    return True

# -*- coding: utf-8 -*-

def get_dict_value(adict:dict, key:str, prefix:list=None):
    if prefix is None:
        prefix = key.split('.')
    if len(prefix) == 1:
        return adict[prefix[0]]
    else:
        return get_dict_value(adict[prefix[0]], key, prefix=prefix[1:])

def get_dict_value_deep(adict:dict, key:str, prefix:list = None, as_array:bool = False, splitter:str = '.'):
    """Used to get value from hierarhic dicts in python with params with dots as splitter"""
    if prefix is None:
        prefix = key.split(splitter)
    if len(prefix) == 1:
        if type(adict) == type({}):
            if not prefix[0] in adict.keys():
                return None
            if as_array:
                return [adict[prefix[0]], ]
            return adict[prefix[0]]
        elif type(adict) == type([]):
            if as_array:
                result = []
                for v in adict:
                    if prefix[0] in v.keys():
                        result.append(v[prefix[0]])
                return result
            else:
                if len(adict) > 0 and prefix[0] in adict[0].keys():
                    return adict[0][prefix[0]]
        return None
    else:
        if type(adict) == type({}):
            if prefix[0] in adict.keys():
                return get_dict_value_deep(adict[prefix[0]], key, prefix=prefix[1:], as_array=as_array)
        elif type(adict) == type([]):
            if as_array:
                result = []
                for v in adict:
                    res = get_dict_value_deep(v[prefix[0]], key, prefix=prefix[1:], as_array=as_array)
                    if res:
                        result.extend(res)
                return result
            else:
                return get_dict_value_deep(adict[0][prefix[0]], key, prefix=prefix[1:], as_array=as_array)
        return None




from __future__ import annotations
import typing
from ..base import BaseCodec

import bz2

class BZIP2Codec(BaseCodec):
    def __init__(self, filename:str, compression_level:int = 5, mode:str = 'r', open_it:bool = False):
        self.compression_level = compression_level
        super(BZIP2Codec, self).__init__(filename, mode=mode, open_it=open_it)

    def open(self) -> bz2.BZ2File:
        self._fileobj = bz2.open(self.filename, self.mode, compresslevel=self.compression_level)
        return self._fileobj

    def close(self):
        self._fileobj.close()

    @staticmethod
    def fileexts() -> list[str]:
        return ['bz2',]



from __future__ import annotations
import typing
from csv import DictReader, DictWriter

from ..base import BaseFileIterable


class CSVIterable(BaseFileIterable):
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, keys: list[str] = None, delimiter:str = ',', quotechar:str='"', mode:str='r', encoding:str = 'utf8'):
        super(CSVIterable, self).__init__(filename, stream, codec=codec, binary=False, encoding=encoding, mode=mode)
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.keys = keys
        self.reset()
        pass

    def reset(self):
        super(CSVIterable, self).reset()
        if self.fobj is None and self.codec is not None:
            fobj = self.codec.textIO(self.encoding)
        else:
            fobj = self.fobj

        self.reader = None
        if self.mode == 'r':
            if self.keys is not None:
                 self.reader = DictReader(fobj, fieldnames=self.keys, delimiter=self.delimiter,
                                     quotechar=self.quotechar)                
            else:
                 self.reader = DictReader(fobj, delimiter=self.delimiter, quotechar=self.quotechar)
        if self.mode in ['w', 'wr'] and self.keys:
            self.writer = DictWriter(fobj, fieldnames=self.keys, delimiter=self.delimiter, quotechar=self.quotechar)
            self.writer.writeheader()
        else:
            self.writer = None

        #            self.reader = reader(self.fobj, delimiter=self.delimiter, quotechar=self.quotechar)
        self.pos = 0

    @staticmethod
    def id() -> str:
        return 'csv'

    @staticmethod
    def is_flatonly() -> bool:
        return True

    def read(self, skip_empty:bool = True):
        """Read single CSV record"""
        row = next(self.reader)
        if skip_empty and len(row) == 0:
            return self.read(skip_empty)
        self.pos += 1
        return row

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk CSV records"""
        chunk = []
        for n in range(0, num):
            chunk.append(next(self.reader))
            self.pos += 1
        return chunk

    def write(self, record:dict):
        """Write single CSV record"""
        self.writer.writerow(record)

    def write_bulk(self, records: list[dict]):
        """Write bulk CSV records"""
        self.writer.writerows(records)



from __future__ import annotations
import typing
from json import loads, dumps
import datetime

from ..base import BaseFileIterable

date_handler = lambda obj: (
    obj.isoformat()
    if isinstance(obj, (datetime.datetime, datetime.date))
    else None
)


class JSONLinesIterable(BaseFileIterable):
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode:str = 'r', encoding:str = 'utf8'):
        super(JSONLinesIterable, self).__init__(filename, stream, codec=codec, binary=False, mode=mode, encoding=encoding)
        self.pos = 0
        pass

    @staticmethod
    def id() -> str:
        return 'jsonl'

    @staticmethod
    def is_flatonly() -> bool:
        return False


    def read(self, skip_empty:bool = False) -> dict:
        """Read single JSON lines record"""
        line = next(self.fobj)
        if skip_empty and len(line) == 0:
            return self.read(skip_empty)
        self.pos += 1
        if line:
            return loads(line)
        return None

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk JSON lines records"""
        chunk = []
        for n in range(0, num):
            chunk.append(loads(self.fobj.readline()))
        return chunk

    def write(self, record: dict):
        """Write single JSON lines record"""
        self.fobj.write(dumps(record, ensure_ascii=False, default=date_handler) + '\n')

    def write_bulk(self, records: list[dict]):
        """Write bulk JSON lines records"""
        for record in records:
            self.fobj.write(dumps(record, ensure_ascii=False, default=date_handler) + '\n')



from ..datatypes.avro import AVROIterable
from ..datatypes.bsonf import BSONIterable
from ..datatypes.csv import CSVIterable
from ..datatypes.orc import ORCIterable
from ..datatypes.parquet import ParquetIterable
from ..datatypes.picklef import PickleIterable
from ..datatypes.json import JSONIterable
from ..datatypes.jsonl import JSONLinesIterable
from ..datatypes.xls import XLSIterable
from ..datatypes.xlsx import XLSXIterable
from ..datatypes.xml import XMLIterable

from ..codecs.bz2codec import BZIP2Codec
from ..codecs.gzipcodec import GZIPCodec
from ..codecs.lzmacodec import LZMACodec
from ..codecs.lz4codec import LZ4Codec
from ..codecs.zipcodec import ZIPCodec

DATATYPES = [AVROIterable, BSONIterable, CSVIterable, ORCIterable, 
             ParquetIterable, PickleIterable, JSONIterable, JSONLinesIterable, 
             XLSIterable, XLSXIterable, XMLIterable]
CODECS = [BZIP2Codec, LZMACodec, GZIPCodec, LZ4Codec, ZIPCodec]

DATATYPE_MAP = {'avro' : AVROIterable, 
                'bson' : BSONIterable, 
                'csv' : CSVIterable,
                'tsv' : CSVIterable,
                'json' : JSONIterable,
                'jsonl' : JSONLinesIterable,
                'ndjson' : JSONLinesIterable,
                'parquet' : ParquetIterable,
                'pickle' : PickleIterable,
                'orc' : ORCIterable,
                'xls' : XLSIterable,
                'xlsx' : XLSXIterable,
                'xml' : XMLIterable
                }

CODECS_MAP = {'bz2' : BZIP2Codec, 
              'gz' : GZIPCodec,
              'lz4' : LZ4Codec,
              'xz' : LZMACodec,
              'lzma' : LZMACodec,
              'zip' : ZIPCodec
              }



def detect_file_type(filename:str) -> dict:
    """Detects file type and compression codec from filename"""
    result = {'filename' : filename, 'success' : False, 'codec' : None, 'datatype' : None}
    parts = filename.lower().split('.')
    if len(parts) == 2:
        if parts[-1] in DATATYPE_MAP.keys():
            result['datatype'] = DATATYPE_MAP[parts[-1]]
            result['success'] = True
    elif len(parts) > 2:
        if parts[-2] in DATATYPE_MAP.keys():
            result['datatype'] = DATATYPE_MAP[parts[-2]]
            result['success'] = True
        if parts[-1] in CODECS_MAP.keys():
            result['codec'] = CODECS_MAP[parts[-1]]
    return result



def open_iterable(filename:str, mode:str = 'r', codecargs:dict={}, iterableargs:dict={}):
    """Opens file and returns iterable object. Codecargs and iterable args are dicts with arguments to codec and iterable class"""
    result = detect_file_type(filename)
    if result['success']:
        if result['codec'] is not None:
            codec = result['codec'](filename=filename, mode=mode, **codecargs)
            iterable = result['datatype'](codec=codec, mode=mode, **iterableargs)
        else:
            iterable = result['datatype'](filename=filename, mode=mode, **iterableargs)
    return iterable



from __future__ import annotations
import typing
import lxml.etree as etree

from .zipped import ZIPSourceWrapper
from ..common.converters import etree_to_dict


class ZIPXMLSource(ZIPSourceWrapper):
    def __init__(self, filename:str = None, tagname:str = None, prefix_strip:bool = True):
        super(ZIPXMLSource, self).__init__(filename)
        self.tagname = tagname
        self.prefix_strip = prefix_strip
        self.reader = etree.iterparse(self.current_file, recover=True)
        pass

    def id(self) -> str:
        return 'zip-xml'

    def is_flat(self) -> bool:
        return False

    def iterfile(self) -> dict:
        res = super(ZIPXMLSource, self).iterfile()
        if res:
            self.reader = etree.iterparse(self.current_file, recover=True)
        return res

    def read_single(self) -> dict:
        """Read single XML record"""
        row = None
        while not row:
            event, elem = next(self.reader)
            shorttag = elem.tag.rsplit('}', 1)[-1]
            if shorttag == self.tagname:
                if self.prefix_strip:
                    row = etree_to_dict(elem, self.prefix_strip)
                else:
                    row = etree_to_dict(elem)
        self.filepos += 1
        self.globalpos += 1
        return row[self.tagname]



from __future__ import annotations
import typing
import avro.schema
from avro.datafile import DataFileReader, DataFileWriter
from avro.io import DatumReader, DatumWriter

from ..base import BaseFileIterable, BaseCodec


class AVROIterable(BaseFileIterable):
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode='r'):
        super(AVROIterable, self).__init__(filename, stream, codec=codec, mode=mode, binary=True)
        self.reset()
        pass

    def reset(self):
        """Reset iterable"""
        super(AVROIterable, self).reset()
        self.pos = 0
        self.cursor = DataFileReader(self.fobj, DatumReader())


    @staticmethod
    def id() -> str:
        return 'avro'


    @staticmethod
    def is_flatonly() -> bool:
        return True


    def read(self) -> dict:
        """Read single record"""
        row = next(self.cursor)
        self.pos += 1
        return row

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk records"""
        chunk = []
        for n in range(0, num):
            chunk.append(self.read())
        return chunk




from __future__ import annotations
import typing
import parquet

from ..base import BaseFileIterable


class ParquetIterable(BaseFileIterable):
    datamode = 'binary'
    def __init__(self, filename:str = None, stream:typing.IO = None, mode: str = 'r', codec: BaseCodec = None):
        super(ParquetIterable, self).__init__(filename, stream, codec=codec, mode=mode, binary=True)
        self.reset()
        pass

    def reset(self):
        """Reset iterable"""
        super(ParquetIterable, self).reset()
        self.pos = 0
        self.cursor = parquet.DictReader(self.fobj)


    @staticmethod
    def id() -> str:
        return 'parquet'

    @staticmethod
    def is_flatonly() -> bool:
        return True


    def read(self) -> dict:
        """Read single Parquet record"""
        row = next(self.cursor)
        self.pos += 1
        return row

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk Parquet records"""
        chunk = []
        for n in range(0, num):
            chunk.append(self.read())
        return chunk




from __future__ import annotations
import typing
from zipfile import ZipFile

from .base import BaseSource


class ZIPSourceWrapper(BaseSource):
    def __init__(self, filename:str, binary:bool = False):
        super(ZIPSourceWrapper, self).__init__()
        self.fobj = ZipFile(filename, mode='r')
        self.filenames = self.fobj.namelist()
        self.filenum = 0
        self.filepos = 0
        self.globalpos = 0
        self.mode = 'rb' if binary else 'r'
        self.current_file = self.fobj.open(self.filenames[self.filenum], mode=self.mode)
        pass

    def close(self):
        if self.current_file:
            self.current_file.close()
            self.current_file = None
        self.fobj.close()

    def iterfile(self) -> bool:
        if self.current_file:
            self.current_file.close()
        if self.filenum < len(self.filenames) - 1:
            self.filenum += 1
            filename = self.filenames[self.filenum]
            self.current_file = self.fobj.open(filename, mode=self.mode)
            self.filepos = 0
            return True
        else:
            return False

    def read(self) -> dict:
        """Read single record"""
        try:
            row = self.read_single()
            return row
        except StopIteration as e:
            if self.iterfile():
                row = self.read_single()
                return row
            else:
                raise StopIteration

    def __iter__(self) -> ZIPSourceWrapper:
        self.filenum = 0
        filename = self.filenames[self.filenum]
        self.current_file = self.fobj.open(filename, mode=self.mode)
        return self

    def read_single(self):
        """Not implemented single record read"""
        raise NotImplementedError

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk records"""
        chunk = []
        n = 0
        while n < num:
            n += 1
            try:
                chunk.append(self.read())
            except StopIteration:
                return chunk
        return chunk



from __future__ import annotations
import typing
import json

from ..base import BaseFileIterable


class JSONIterable(BaseFileIterable):
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode='r', tagname:str= None):
        super(JSONIterable, self).__init__(filename, stream, codec=codec, mode=mode, binary=False)
        self.tagname = tagname
        self.reset()
        pass

    def reset(self):
        super(JSONIterable, self).reset()
        self.pos = 0
        self.data = json.load(self.fobj)
        if self.tagname:
            self.data = self.data[self.tagname]
        self.total = len(self.data)

    @staticmethod
    def id() -> str:
        return 'json'

    @staticmethod
    def is_flatonly() -> bool:
        return False

    def read(self, skip_empty:bool = False) -> dict:
        """Read single JSON record"""
        if self.pos >= self.total:
            raise StopIteration

        row = self.data[self.pos]
        self.pos += 1
        return row

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk JSON records"""
        chunk = []
        for n in range(0, num):
            chunk.append(self.read())
        return chunk



from .avro import AVROIterable
from .bsonf import BSONIterable
from .csv import CSVIterable
from .orc import ORCIterable
from .parquet import ParquetIterable
from .picklef import PickleIterable
from .json import JSONIterable
from .jsonl import JSONLinesIterable
from .xls import XLSIterable
from .xlsx import XLSXIterable
from .xml import XMLIterable



from __future__ import annotations
import typing
from openpyxl import load_workbook

from ..base import BaseFileIterable

def read_row_keys(rownum, ncols, sheet):
    """Read single row by row num"""
    tmp = list()
    for i in range(0, ncols):
        ct = sheet.cell_type(rownum, i)
        cell_value = sheet.cell_value(rownum, i)
        get_col = str(cell_value)
        tmp.append(get_col)    
    return tmp


class XLSXIterable(BaseFileIterable):
    datamode = 'binary'
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode='r', keys: list[str] = None, page:int = 0, start_line:int = 0):
        super(XLSXIterable, self).__init__(filename, stream, codec=codec, mode=mode, binary=True, noopen=True)
        self.keys = keys
        self.start_line = start_line + 1
        self.page = page
        self.pos = self.start_line
        self.extracted_keys = keys is None
        self.reset()
        pass

    def reset(self):
        super(XLSXIterable, self).reset()
        self.workbook = load_workbook(self.filename)
        self.sheet = self.workbook.active
        self.pos = self.start_line
        self.cursor = self.sheet.iter_rows()
        if self.pos > 1:
            self.skip(self.pos - 1)
        if self.extracted_keys:
            self.keys = list()
            row = next(self.cursor)
            for cell in row:
                self.keys.append(str(cell.value))             
            self.pos += 1

    def skip(self, num:int = 1):
        while num > 0:
            num -= 1
            o = next(self.cursor)

    
    @staticmethod
    def id() -> str:
        return 'xlsx'

    @staticmethod
    def is_flatonly() -> bool:
        return True

    def read(self) -> dict:
        """Read single XLSX record"""
        row = next(self.cursor)
        tmp = list()
        for cell in row:
            tmp.append(str(cell.value))
        result = dict(zip(self.keys, tmp))
        self.pos += 1
        return result

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk XLSX records"""
        chunk = []
        for n in range(0, num):
            row = next(self.cursor)
            tmp = list()
            for cell in row:
                tmp.append(str(cell.value))
            result = dict(zip(self.keys, tmp))
            chunk.append(result)
            self.pos += 1
        return chunk



from __future__ import annotations
import typing
import pickle
import datetime

from ..base import BaseFileIterable

date_handler = lambda obj: (
    obj.isoformat()
    if isinstance(obj, (datetime.datetime, datetime.date))
    else None
)


class PickleIterable(BaseFileIterable):
    datamode = 'binary'
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode:str = 'r'):
        super(PickleIterable, self).__init__(filename, stream, codec=codec, binary=True, mode=mode)
        self.pos = 0
        pass

    @staticmethod
    def id() -> str:
        return 'pickle'

    @staticmethod
    def is_flatonly() -> bool:
        return False

    def read(self) -> dict:
        """Read single record"""
        try:
            return pickle.load(self.fobj)
        except EOFError:
            raise StopIteration
	

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk records"""
        chunk = []
        for n in range(0, num):
            try:
                obj = pickle.load(self.fobj)            
                chunk.append(obj)
            except:
                if len(chunk) > 0: return chunk
                raise StopIteration
        return chunk


    def write(self, record: dict):
        """Write single record into file"""
        pickle.dump(record, self.fobj)

    def write_bulk(self, records: list[dict]):
        """Write bulk records"""
        for record in records:
            pickle.dump(record, self.fobj)



from __future__ import annotations
import typing
from ..base import BaseCodec

import lzma

LZMA_FILTERS = [
    {"id": lzma.FILTER_DELTA, "dist": 5},
    {"id": lzma.FILTER_LZMA2, "preset": 7 | lzma.PRESET_EXTREME},
]

class LZMACodec(BaseCodec):
    def __init__(self, filename:str, compression_level:int = 5, mode:str = 'r', open_it:bool = False):
        self.compression_level = compression_level
        super(LZMACodec, self).__init__(filename, mode=mode, open_it=open_it)

    def open(self) -> lzma.LZMAFile:
        filters = LZMA_FILTERS
        filters[0]['dist'] = self.compression_level
        self._fileobj = lzma.LZMAFile(self.filename, mode=self.mode, format=lzma.FORMAT_XZ)#, filters=filters)
        return self._fileobj


    def reset(self):
        if self.mode in ['w', 'wb']:
            pass
        else:
            super(LZMACodec, self).reset()

    def close(self):
        self._fileobj.close()

        
    @staticmethod
    def fileexts() -> list[str]:
        return ['xz', 'lzma']



# This is purely the result of trial and error.

import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

import iterable


class PyTest(TestCommand):
    # `$ python setup.py test' simply installs minimal requirements
    # and runs the tests with no fancy stuff like parallel execution.
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--doctest-modules', '--verbose',
            './tests'
        ]
        self.test_suite = True

    def run_tests(self):
        import pytest
        sys.exit(pytest.main(self.test_args))


tests_require = [
    # Pytest needs to come last.
    # https://bitbucket.org/pypa/setuptools/issue/196/
    'pytest',
    'mock',
]

install_requires = [
    'xlrd',
    'pyorc',
    'parquet',
    'openpyxl',
    'jsonlines',
    'orjson',
    'lz4',
    'chardet',
    'avro',
    'lxml'
]

# Conditional dependencies:

# sdist
if 'bdist_wheel' not in sys.argv:
    try:
        # noinspection PyUnresolvedReferences
        import argparse
    except ImportError:
        install_requires.append('argparse>=1.2.1')


# bdist_wheel
extras_require = {
    # https://wheel.readthedocs.io/en/latest/#defining-conditional-dependencies
    'python_version == "3.8" or python_version == "3.8"': ['argparse>=1.2.1'],
}



setup(
    name='iterabledata',
    version=iterable.__version__,
    description=iterable.__doc__.strip(),
    long_description=open('README.md', 'r', encoding='utf8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/apicrafter/pyiterable/',
    download_url='https://github.com/apicrafter/pyiterable/',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    author=iterable.__author__,
    author_email='ivan@begtin.tech',
    license=iterable.__licence__,
    entry_points={},
    extras_require=extras_require,
    install_requires=install_requires,
    tests_require=tests_require,
    python_requires='>=3.8',
    cmdclass={'test': PyTest},
    zip_safe=False,
    keywords='json jsonl csv bson parquet orc xml xls xlsx dataset etl data-pipelines',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development',
        'Topic :: Text Processing',
    ],
)



from __future__ import annotations
import typing
import pyorc

from ..base import BaseFileIterable

def df_to_pyorc_schema(df):
    """Extracts column information from pandas dataframe and generate pyorc schema"""
    struct_schema = []
    for k, v in df.dtypes.to_dict().items():
        v = str(v)
        if v == 'float64':
            struct_schema.append('%s:float' % (k))
        elif v == 'float32':
            struct_schema.append('%s:float' % (k))
        elif v == 'datetime64[ns]':
            struct_schema.append('%s:timestamp' % (k))
        elif v == 'int32':
            struct_schema.append('%s:int' % (k))
        elif v == 'int64':
            struct_schema.append('%s:int' % (k))
        else:
            struct_schema.append('%s:string' %(k))
    return struct_schema


def fields_to_pyorc_schema(fields):
    """Converts list of fields to pyorc schema array"""
    struct_schema = []
    for field in fields:
        struct_schema.append('%s:string' %(field))    
    return struct_schema

class ORCIterable(BaseFileIterable):
    datamode = 'binary'
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode:str = 'r', keys:list[str] = None, schema:list[str] = None, compression:int = 5):
        self.keys = keys
        self.schema = schema
        self.compression = compression
        super(ORCIterable, self).__init__(filename, stream, codec=codec, mode=mode, binary=True)
        self.reset()
        pass

    def reset(self):
        """Reset iterable"""
        super(ORCIterable, self).reset()
        self.pos = 0
        self.reader = None
        if self.mode == 'r':
            self.reader = pyorc.Reader(self.fobj, struct_repr=pyorc.StructRepr.DICT)
        self.writer = None
        if self.mode == 'w':
            if self.schema is not None:
               struct_schema = self.schema
            else:
               struct_schema = fields_to_pyorc_schema(self.keys)
            self.writer = pyorc.Writer(self.fobj, "struct<%s>" % (','.join(struct_schema)), struct_repr = pyorc.StructRepr.DICT, compression=self.compression, compression_strategy=1)  
         

    @staticmethod
    def id() -> str:
        return 'orc'

    @staticmethod
    def is_flatonly() -> bool:
        return True

    def close(self):
        """Close iterable"""
        if self.writer is not None: self.writer.close()
        super(ORCIterable, self).close()


    def read(self) -> dict:
        """Read single record"""
        row = next(self.reader)
        self.pos += 1
        return row

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk records"""
        chunk = []
        for n in range(0, num):
            chunk.append(self.read())
        return chunk

    def write(self, record: dict):
        """Write single record"""
        self.writer.write(record)

    def write_bulk(self, record: list[dict]):
        """Write bulk records"""
        self.writer.writerows(records)



from __future__ import annotations
import typing
from ..base import BaseCodec


class RAWCodec(BaseCodec):
    def __init__(self, filename:str, mode:str = 'r'):
        super(RAWCodec, self).__init__(filename, mode=mode, open_it=open_it)

    def open(self) -> typing.IO:
        self._fileobj = open(self.filename, self.mode)
        return self._fileobj

    def close(self):
        self._fileobj.close()

    @staticmethod
    def fileexts() -> list[str]:
        return None



from __future__ import annotations
import typing
from collections import defaultdict
import lxml.etree as etree

from ..base import BaseFileIterable


PREFIX_STRIP = False
PREFIX = ""


def etree_to_dict(t, prefix_strip=True):
    """Converts tree of XML elements from lxml to python dictionary"""
    tag = t.tag if not prefix_strip else t.tag.rsplit('}', 1)[-1]
    d = {tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            #            print(dir(dc))
            for k, v in dc.items():
                if prefix_strip:
                    k = k.rsplit('}', 1)[-1]
                dd[k].append(v)
        d = {tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[tag].update(('@' + k.rsplit('}', 1)[-1], v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            tag = tag.rsplit('}', 1)[-1]
            if text:
                d[tag]['#text'] = text
        else:
            d[tag] = text
    return d


class XMLIterable(BaseFileIterable):
    datamode = 'binary'
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode='r', tagname:str = None, prefix_strip:bool = True):
        super(XMLIterable, self).__init__(filename, stream, codec=codec, mode=mode, binary=True, encoding='utf8')
        self.tagname = tagname
        self.prefix_strip = prefix_strip
        self.reset()
        pass

    def reset(self):
        super(XMLIterable, self).reset()
        self.reader = etree.iterparse(self.fobj, recover=True)        
        self.pos = 0

    @staticmethod
    def id() -> str:
        return 'xml'


    @staticmethod
    def is_flatonly() -> bool:
        return False

    def read(self) -> dict:
        """Read single XML record"""
        row = None
        while not row:
            event, elem = next(self.reader)
            shorttag = elem.tag.rsplit('}', 1)[-1]
            if shorttag == self.tagname:
                if self.prefix_strip:
                    row = etree_to_dict(elem, self.prefix_strip)
                else:
                    row = etree_to_dict(elem)
        self.pos += 1
        return row[self.tagname]

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk XML records"""
        chunk = []
        for n in range(0, num):
            chunk.append(self.read())
        return chunk



from __future__ import annotations
import typing
from ..base import BaseCodec

import zipfile

class ZIPCodec(BaseCodec):
    def __init__(self, filename:str, compression_level:int = 5, mode:str = 'r', open_it:bool = False):
        self.compression_level = compression_level            
        if mode == 'rb':
            mode = 'r'
            self.filemode = 'rb'        
        else:
            self.filemode = 'r'
        super(ZIPCodec, self).__init__(filename, mode=mode, open_it=open_it)

    def open(self) -> zipfile.ZipFile:        
        self._archiveobj = zipfile.ZipFile(self.filename, mode=self.mode)
        fnames = self._archiveobj.namelist()
        self._fileobj = self._archiveobj.open(fnames[0], self.filemode)
        return self._fileobj

    def close(self):
        self._fileobj.close()
        self._archiveobj.close()

    @staticmethod
    def fileexts() -> list[str]:
        return ['zip',]



from __future__ import annotations
import typing
from ..base import BaseCodec

import gzip

class GZIPCodec(BaseCodec):
    def __init__(self, filename:str, compression_level:int = 5, mode:str = 'r', open_it:bool = False):
        self.compression_level = compression_level
        super(GZIPCodec, self).__init__(filename, mode=mode, open_it=open_it)

    def open(self) -> gzip.GzipFile:         
        self._fileobj = gzip.GzipFile(filename=self.filename, mode=self.mode, compresslevel=self.compression_level)
        return self._fileobj
 
    def close(self):
        self._fileobj.close()

    @staticmethod
    def fileexts() -> list[str]:
        return ['gz',]



# from bson import BSON
from __future__ import annotations
import typing
import bson

from ..base import BaseFileIterable, BaseCodec


class BSONIterable(BaseFileIterable):
    datamode = 'binary'
    def __init__(self, filename:str = None, stream:typing.IO = None, codec: BaseCodec = None, mode:str='r'):
        super(BSONIterable, self).__init__(filename, stream, codec=codec, binary=True, mode=mode)
        self.reset()
        pass

    def reset(self):
        super(BSONIterable, self).reset()
        self.reader = bson.decode_file_iter(self.fobj)

    @staticmethod
    def id() -> str:
        return 'bson'


    @staticmethod
    def is_flatonly() -> bool:
        return False

    def read(self) -> dict:
        """Write single bson record"""
        return next(self.reader)

    def read_bulk(self, num:int = 10) -> list[dict]:
        """Read bulk bson record"""
        chunk = []
        for n in range(0, num):
            chunk.append(next(self.reader))
        return chunk

    def write(self, record:dict):
        """Write single bson record"""
        self.fobj.write(bson.BSON.encode(record))

    def write_bulk(self, records:list[dict]):
        """Write bulk bson record"""
        for record in records:
            self.fobj.write(bson.BSON.encode(record))



# -*- coding: utf-8 -*-
from ..base import BaseIterable
from typing import Callable
import time


def pipeline(source: BaseIterable, destination: BaseIterable, process_func: Callable[[dict, dict], dict], trigger_func:Callable[[dict, dict], None] = None, trigger_on:int=1000, final_func: Callable[[dict, dict], None] = None, reset_iterables: bool = True, skip_nulls: bool = True, start_state:dict={}, debug:bool = False):
    """Wrapper over Pipeline class to simplify data processing pipelines execution"""
    runner = Pipeline(source=source, destination=destination, process_func=process_func, trigger_func=trigger_func, trigger_on=trigger_on, final_func=final_func, reset_iterables=reset_iterables, skip_nulls=skip_nulls, start_state=start_state)
    runner.run(debug)


class Pipeline:
    """Data processing pipeline that read data and process it"""
    def __init__(self, source: BaseIterable, destination: BaseIterable = None, process_func: Callable[[dict, dict], dict]=None, trigger_func:Callable[[dict, dict], None] = None, trigger_on:int=1000, final_func: Callable[[dict, dict], None] = None, reset_iterables: bool = True, skip_nulls: bool = True, start_state:dict={}):
        self.source = source
        self.destination = destination
        self.process_func = process_func
        self.trigger_func = trigger_func
        self.trigger_on = trigger_on
        self.final_func = final_func
        self.reset_iterables = reset_iterables
        self.skip_nulls = skip_nulls
        self.start_state={}
        pass


    def run(self, debug:bool = False):
        """Execute pipeline"""
        stats = {'rec_count' : 0, 'nulls' : 0, 'exceptions' : 0, 'time_start' : time.time()}    
        state = self.start_state
        if self.reset_iterables:
            self.source.reset()
            if self.destination is not None:
                self.destination.reset()
        
        for record in self.source:
            try:
                result = self.process_func(record, state)
                if result is None:
                    if not self.skip_nulls:
                        stats['nulls'] += 1
                        if self.destination is not None:
                            self.destination.write(result)                
                else:
                    self.destination.write(result)                
            except Exception as e:
                print(e)
                stats['exceptions'] += 1
                pass
            stats['rec_count'] += 1
            if stats['rec_count'] % self.trigger_on == 0 and self.trigger_func is not None:
                try:
                    self.trigger_func(stats, state)
                except Exception as e:
# TODO: Add better error handling here using logging
                    pass
        stats['time_end'] = time.time()
        stats['duration'] = stats['time_end'] - stats['time_start']
        if self.final_func is not None:
            self.final_func(stats, state)
        return stats